from typing import Dict, Sequence
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
from virocon import get_OMAE2020_Hs_Tz, GlobalHierarchicalModel,IFORMContour
from xclim.indices.stats import frequency_analysis
"""
Temporarily copied from https://github.com/MET-OM/metocean-stats/blob/main/metocean_stats/stats/extreme_stats.py with permissions

"""

def return_value(hs: xr.DataArray,return_periods=[50,100], month=[1,2,3,4,5, 6, 7, 8, 9, 10,11,12], image_path=None) -> xr.DataArray:
    time = hs.time
    r_val = frequency_analysis(hs, t=return_periods, dist="genextreme", mode="max", freq="YS", month=month)
    if image_path:
        color = plt.cm.rainbow(np.linspace(0, 1, len(r_val.return_period)))
        _, ax = plt.subplots()
        hs.plot(color='grey')
        for i in range(len(r_val.return_period)):
            lbl = str(r_val.values[i].round(1))+r_val.units+'('+str(r_val.return_period[i].values)+'y)'
            ax.hlines(y=r_val[i], xmin=time[0], xmax=time[-1],color=color[i] ,linestyle='dashed', linewidth=2,label=lbl )
        plt.grid()
        plt.legend(loc='center left')
        plt.title('Return value')
        plt.tight_layout()
        plt.savefig(image_path)

    return r_val

def joint_2D_contour(hs: xr.DataArray,tp: xr.DataArray, return_periods=[50,100], image_path=None, image_title: str = None) -> Sequence[Dict]:
    """Compute a joint contour for the given return periods. Optionally plot the contour and save the image.
    Input:
        hs: Hs values
        tp: Tp values
        return_periods: A list of return periods in years
        image_path: Path to save the image (optional)
        image_title: Title of the image (optional)
    """
    # Define joint distribution model
    dist_descriptions, fit_descriptions, _ = get_OMAE2020_Hs_Tz()
    model = GlobalHierarchicalModel(dist_descriptions)
    # Perform fitting
    data =  np.transpose(np.array([hs,tp]))
    model.fit(data, fit_descriptions=fit_descriptions)
    # Create a list of contours for each return period
    dt = 1  # duration in hours
    contours = []
    for rp in return_periods:
        alpha = 1 / (rp * 365.25 * 24 / dt)
        contour = IFORMContour(model, alpha)
        coords = contour.coordinates
        x = coords[:, 1].tolist()
        y = coords[:, 0].tolist()
        contours.append({
            'return_period': rp,
            'x': x,
            'y': y
        })

    # Optionally plot the contours and save the image
    if image_path:
        import seaborn as sns
        _, ax = plt.subplots()
        sns.set_theme(style="ticks")
        sns.scatterplot(x=tp, y=hs, ax=ax)
        color = plt.cm.rainbow(np.linspace(0, 1, len(return_periods)))
        # Compute an IFORM contour with a return period of N years
        dt = 1  # duration in hours
        #loop over contours and also get index

        for i,contour in enumerate(contours):
            # Plot the contour
            x = []
            x.extend(contour['x'])
            x.append(x[0])

            y = []
            y.extend(contour['y'])
            y.append(y[0])

            rp = contour['return_period']
            ax.plot(x, y, color=color[i],label=str(rp)+'y')

        if hs.name and tp.name:
            ax.set_xlabel(tp.name)
            ax.set_ylabel(hs.name)
        plt.grid()
        if image_title:
            plt.title(image_title)
        plt.legend()
        plt.savefig(image_path)

    return contours



from datetime import datetime
from typing import Sequence

from dateutil import rrule

from bluesmet.normet.dataset import Dataset

from ..combine import get_values_between as get_values


def get_values_between(
    lat_pos: float,
    lon_pos: float,
    start_date: datetime,
    end_date: datetime,
    requested_values: Sequence[str],
    cache_location="./cache/arome3km_3hr/",
    max_concurrent_downloads = 1
):
    """Return values for Arome 3k atmosphere subset"""

    return get_values(
        requested_values,
        lat_pos,
        lon_pos,
        "longitude",
        "latitude",
        start_date,
        end_date,
        rrule.MONTHLY,
        cache_location,
        __get_url,
        max_concurrent_downloads
    )


def __get_url(date: datetime):
    basename = "arome3km_3hr"
    datlab = date.strftime("%Y%m")
    base_url = "https://thredds.met.no/thredds/dodsC/nora3_subset_atmos/atm_3hourly"
    return f"{base_url}/{basename}_{datlab}.nc"


def get_metadata():
    """Get metadata"""
    from_date = datetime(1970, 1, 1)
    to_date = datetime(2023, 7, 31)
    url = __get_url(from_date)
    ds = Dataset("./cache/arome3km_3hr/", url, from_date)
    metadata = ds.get_metadata()
    metadata["global"] = {
        "fromDate": from_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
        "toDate": to_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
        "url": "https://thredds.met.no/thredds/catalog/nora3_subset_atmos/atm_3hourly/catalog.html",
    }
    return metadata


import numpy as np
from scipy.spatial import cKDTree

# def calculate_true_lon_lat(ds):
#     dvars = ds.variables
#     rlon = dvars["rlon"]
#     rlat = dvars["rlat"]
#     proj = dvars["projection_ob_tran"]
#     rlon, rlat = np.meshgrid(rlon, rlat, sparse=False, indexing="ij")
#     tf = pyproj.Transformer.from_crs(proj.proj4, "epsg:4326", always_xy=True)
#     # pylint: disable=unpacking-non-sequence
#     lon, lat = tf.transform(rlon, rlat)
#     return lon, lat



def nearest(lat,lon, lat_pos, lon_pos):
    """Function to find index to nearest point """
    M = np.c_[np.ravel(lon), np.ravel(lat)]
    tree = cKDTree(M)
    _, ii = tree.query([lon_pos, lat_pos], k=1)
    idy, idx = np.where((lon == M[ii][0]) & (lat == M[ii][1]))
    return int(idx), int(idy)



import os
from pathlib import Path
import shutil
from datetime import datetime
from typing import Dict, Sequence

import netCDF4 as nc
import numpy as np
from numpy import ndarray
from scipy.spatial import cKDTree

from .cache import Cache


class Dataset:
    """Met.no dataset for a given date"""

    def __init__(self, cache_location: str, url: str, date: datetime):
        self.cache = Cache(location=cache_location)
        self.url = url
        self.date = date

    def get_actual_coordinates(
        self,
        lat_pos: float,
        lon_pos: float,
        lon_name="longitude",
        lat_name="latitude",
    ) -> Dict:
        """Find actual coordinates and dimensions to reduce on"""
        coordinates = {}
        latlondims = {}
        root_vars = self.cache.get_root_variables(self.url, [lon_name, lat_name], latlondims)

        lons = root_vars[lon_name]
        lats = root_vars[lat_name]

        dim1, dim2, lon_a, lat_a = Dataset.__find_nearest(lons, lats, lon_pos, lat_pos)
        dims = latlondims[lon_name]
        coordinates["longitude_actual"] = lon_a
        coordinates["latitude_actual"] = lat_a

        coordinates["dim1"] = {dims[0]: dim1}
        coordinates["dim2"] = {dims[1]: dim2}
        return coordinates

    def __update_coordinates(
        self,
        coordinates: Dict
    ) -> Dict:
        """Find actual coordinates and dimensions to reduce on"""
        lon_pos = coordinates["longitude"]
        lat_pos = coordinates["latitude"]
        lon_name = coordinates.get("lon_name", "longitude")
        lat_name = coordinates.get("lat_name", "latitude")
        latlondims = {}
        root_vars = self.cache.get_root_variables(self.url, [lon_name, lat_name], latlondims)

        lons = root_vars[lon_name]
        lats = root_vars[lat_name]

        dim1, dim2, lon_a, lat_a = Dataset.__find_nearest(lons, lats, lon_pos, lat_pos)
        dims = latlondims[lon_name]
        coordinates["longitude_actual"] = lon_a
        coordinates["latitude_actual"] = lat_a

        coordinates["dim1"] = {dims[0]: dim1}
        coordinates["dim2"] = {dims[1]: dim2}
        return coordinates

    def get_root_values(self, values: dict):
        """Get root variables based on dictionary keys"""
        names = values.keys()
        variables = self.cache.get_root_variables(self.url, names, dimensions={})
        for name, var in zip(names, variables):
            values[name] = var

    def get_values(
        self,
        coordinates: Dict,
        requested_values: Sequence[str],
        dimensions: Dict,
    ) -> Path:
        """Get  data organized using lat,lon and date. Will cache data. Returns path to cached file"""
        new_values = {"time": None}
        for name in requested_values:
            new_values[name] = None

        if "dim1" not in coordinates:
            # Then we need to read the root variables
            self.__update_coordinates(coordinates)

        dim1 = coordinates["dim1"]
        dim2 = coordinates["dim2"]

        reduction_dims = {}
        reduction_dims.update(dim1)
        reduction_dims.update(dim2)

        date_cache_path = self.cache.to_cache_path(self.date, dim1, dim2)

        if not os.path.exists(date_cache_path):
            # Then we need to read the data into the cache
            self.__download(
                self.url,
                requested_values,
                new_values,
                reduction_dims,
                date_cache_path,
                dimensions,
            )
        else:
            missing = {}
            with nc.Dataset(date_cache_path, "r") as fcache:
                for key in new_values:
                    if key not in fcache.variables:
                        missing[key] = None
                    else:
                        var = fcache.variables[key]
                        new_values[key] = var[:]
                        dimensions[key] = var.dimensions
            if len(missing) > 0:
                self.__append(self.url, missing, reduction_dims, date_cache_path, dimensions)
                new_values.update(missing)

        return date_cache_path

    def __reduce_dims(self, var, reduction_dims):
        newdims = {}
        shape = var.shape
        dims = var.dimensions
        for dim in dims - reduction_dims.keys():
            newdims[dim] = range(0, shape[dims.index(dim)])
        for dim in reduction_dims:
            newdims[dim] = reduction_dims[dim]
        thedims = [newdims[dim] for dim in dims]
        return var[tuple(thedims)]

    def __download(
        self,
        url,
        requested_values,
        new_values,
        reduction_dims,
        date_cache_path,
        dimensions,
    ):
        with nc.Dataset(url, "r") as fonline:
            print(f"Downloading {url}")

            parent_dir = date_cache_path.parent
            if not parent_dir.exists():
                os.makedirs(parent_dir)

            time = fonline.variables["time"][:]
            new_values["time"] = time

            filedims = fonline.dimensions
            for dim in reduction_dims:
                if dim not in filedims:
                    raise ValueError(f"Reducing dimension {dim} not found in file {url}")

            variables = fonline.variables
            intermediate = date_cache_path.with_suffix(".tmp")

            if date_cache_path.exists():
                shutil.copy(date_cache_path, intermediate)

            try:
                with nc.Dataset(intermediate, "w", format="NETCDF4") as dst:
                    remaining_dims = filedims.keys() - reduction_dims.keys()
                    for name, dim in filedims.items():
                        if name in remaining_dims:
                            dst.createDimension(name, dim.size)
                    var_names = requested_values
                    if "time" not in var_names:
                        var_names += ["time"]
                    for var_name in var_names:
                        var = variables[var_name]
                        vals = self.__reduce_dims(var, reduction_dims)
                        var_dims = [
                            dim for dim in var.dimensions if dim not in reduction_dims
                        ]
                        new_var = dst.createVariable(var_name, var.dtype, var_dims)
                        new_var.setncatts(var.__dict__)
                        new_var[:] = vals
                        new_values[var_name] = vals
                        dimensions[var_name] = var_dims
                shutil.move(intermediate, date_cache_path)
            except Exception as e:
                if intermediate.exists():
                    os.remove(intermediate)
                raise e

    def __append(self, url, missing, reduction_dims, date_cache_path, dimensions):
        with nc.Dataset(url, "r") as fonline:
            print(f"Downloading {url}")

            intermediate = date_cache_path.with_suffix(".tmp")

            if date_cache_path.exists():
                shutil.copy(date_cache_path, intermediate)

            try:
                with nc.Dataset(intermediate, "a", format="NETCDF4") as dst:
                    for var_name in missing.keys():
                        var = fonline.variables[var_name]
                        vals = self.__reduce_dims(var, reduction_dims)
                        var_dims = [
                            dim for dim in var.dimensions if dim not in reduction_dims
                        ]
                        new_var = dst.createVariable(var_name, var.dtype, var_dims)
                        new_var.setncatts(var.__dict__)
                        new_var[:] = vals
                        missing[var_name] = vals
                        dimensions[var_name] = var_dims
                shutil.move(intermediate, date_cache_path)
            except Exception as e:
                if intermediate.exists():
                    os.remove(intermediate)
                raise e


    def get_metadata(self):
        """Get metadata"""
        # TODO: USE CACHE
        metadata = {}
        with nc.Dataset(self.url, "r") as f0:
            metadata["global"] = dict(f0.__dict__)
            variables = {}
            metadata["variables"] = variables
            for name, var in f0.variables.items():
                var_meta = {}
                variables[name] = var_meta
                for key, value in var.__dict__.items():
                    if key.startswith("_"):
                        continue
                    if isinstance(value, ndarray):
                        var_meta[key] = str(value.tolist())
                    elif isinstance(value, str):
                        var_meta[key] = value
                    else:
                        var_meta[key] = str(value)
                dims = var.dimensions
                if len(dims) > 0:
                    if len(dims) == 1 and dims[0] == name:
                        continue
                    var_meta["dimensions"] = ",".join(dims)
        return metadata

    @staticmethod
    def __nearest(lon, lat, lon_pos, lat_pos):
        """Find nearest point in lon lat matrix"""
        M = np.c_[np.ravel(lon), np.ravel(lat)]
        tree = cKDTree(M)
        _, ii = tree.query([lon_pos, lat_pos], k=1)
        d1, d2 = np.where((lon == M[ii][0]) & (lat == M[ii][1]))
        return int(d1), int(d2)

    @staticmethod
    def __find_nearest(lons, lats, lon_pos, lat_pos):
        dim1, dim2 = Dataset.__nearest(lons, lats, lon_pos, lat_pos)
        lon_actual = lons[dim1, dim2]
        lat_actual = lats[dim1, dim2]
        return dim1, dim2, lon_actual, lat_actual


    @staticmethod
    def __concatenate_time_arrays(values, new, dimensions):
        for name, vals in new.items():
            value = values.get(name)
            if value is None:
                values[name] = vals
            elif "time" in dimensions[name]:
                cced = np.ma.concatenate((value, vals))
                values[name] = cced



import json
from math import floor
from typing import Callable, Dict, Sequence, Tuple

import numpy as np
from numpy import arange, ndarray, number


class Scatter:
    """A class for storing and combining scatter data
    """

    class Bins(dict):
        """Based on collections.Counter but with support for adding lists"""

        def __add__(self, other):
            if not isinstance(other, Scatter.Bins):
                return NotImplemented
            result = Scatter.Bins()
            for elem, count in self.items():
                ov = other[elem]
                newcount = count + ov
                result[elem] = newcount
            for elem, count in other.items():
                if elem not in self:
                    result[elem] = count
            return result

    def __init__(self, bin_size=1.0) -> None:
        """
        Parameters
        ----------
        bin_size : float
            The bin size to use default is 1.0
        """
        self.bin_size = bin_size
        self.bins = Scatter.Bins()

    def add(self, x: number, y: number, **kwargs) -> Dict:
        """Add a point to the scatter. Optionally add additional data stored in the bin"""
        xbin = floor(x / self.bin_size)
        ybin = floor(y / self.bin_size)
        ybins = self.bins.get(xbin, Scatter.Bins())
        cbin = ybins.get(ybin, Scatter.Bins())
        cbin["occurences"] = cbin.get("occurences", 0) + 1
        for key, value in kwargs.items():
            existing = cbin.get(key, [])
            existing.append(value)
            cbin[key] = existing
        ybins[ybin] = cbin
        self.bins[xbin] = ybins
        return cbin

    def combine(self, other: "Scatter"):
        """Combine with given scatter"""
        if self.bin_size != other.bin_size:
            raise ValueError("Bin sizes differ")
        xbins1 = self.bins
        xbins2 = other.bins
        for (idx, counter) in xbins2.items():
            existing = xbins1.get(idx)
            if existing:
                # Add the counts
                for (ybin, count) in existing.items():
                    newcount = counter.get(ybin)
                    if newcount:
                        existing[ybin] = count + newcount

                for (ybin, count) in counter.items():
                    if ybin not in existing:
                        existing[ybin] = count
            else:
                # New bin
                xbins1[idx] = counter

    def num_rows(self):
        """Number of rows"""
        bins = [int(row) for row in self.bins.keys()]
        return max(bins) + 1

    def num_columns(self):
        """Number of columns"""
        largest = -1
        for ybin in self.bins.values():
            bins = [int(row) for row in ybin.keys()]
            largest = max(largest, max(bins))
        return largest + 1

    def row_values(self) -> ndarray:
        """Upper row values"""
        nrows = self.num_rows() + 1
        return arange(self.bin_size, nrows * self.bin_size, self.bin_size)

    def column_values(self) -> ndarray:
        """Upper column values"""
        ncols = self.num_columns() + 1
        return arange(self.bin_size, ncols * self.bin_size, self.bin_size)

    def occurences(self) -> ndarray:
        """Occurences of all the bins"""
        nrows = self.num_rows()
        ncols = self.num_columns()
        occurences = np.zeros((nrows, ncols), dtype=int)
        for (row, ybins) in self.bins.items():
            for (column, ybin) in ybins.items():
                occurences[row][column] = int(ybin["occurences"])
        return occurences

    def mean_xy(self) -> Tuple[number, number]:
        """Mean x and y"""
        xp = 0.0
        yp = 0.0
        for (row, ybins) in self.bins.items():
            for (column, ybin) in ybins.items():
                occ = float(ybin["occurences"])
                x = (row + 0.5) * self.bin_size
                y = (column + 0.5) * self.bin_size
                xp += occ * x
                yp += occ * y
        all_occurences = float(self.occurences().sum())
        xp = xp / all_occurences
        yp = yp / all_occurences
        return (xp, yp)

    def get_values(self, key, reducer: Callable[[Sequence],float]=None) -> ndarray:
        """Get the values for a given key, reduse the list of values using the reducer or the mean if no reducer is given"""
        nrows = self.num_rows()
        ncols = self.num_columns()
        values = np.zeros((nrows, ncols), dtype=float)
        for (row, ybins) in self.bins.items():
            for (column, ybin) in ybins.items():
                kvalues = ybin.get(key)
                if kvalues:
                    if reducer:
                        values[row][column] = reducer(kvalues)
                    else:
                        values[row][column] = sum(kvalues) / len(kvalues)
        return values

    def __str__(self):
        return json.dumps(
            {
                "row": self.row_values().tolist(),
                "column": self.column_values().tolist(),
                "occurences": self.occurences().tolist(),
            },
        )


def main():
    """Demonstrate the scatter"""
    bin_size = 4.0
    scatter1 = Scatter(bin_size)
    scatter1.add(10.0, 0.1, wavedir=10.0)
    scatter1.add(11.0, 0.2, wavedir=11.0)
    scatter1.add(10.0, 1.0, wavedir=12.0)
    scatter1.add(10.0, 2.0, wavedir=12.0)
    scatter1.add(10.1, 10.0, wavedir=13.0)

    scatter2 = Scatter(bin_size)
    scatter2.add(10.0, 0.1, wavedir=15.0)
    scatter2.add(12.0, 0.2, wavedir=15.0)
    scatter2.add(11.0, 1.0, wavedir=20.0)
    scatter2.add(11.0, 2.0, wavedir=17.0)
    scatter2.add(10.1, 10.0, wavedir=9.0)

    print("scatter1=")
    print(scatter1)
    print("scatter2=")
    print(scatter2)
    print("combined")
    scatter1.combine(scatter2)
    print(scatter1)

    print("hs=" + str(scatter1.row_values()))
    print("tp=" + str(scatter1.column_values()))
    print(scatter1.occurences())
    print("Scatter of wavedir=")


    print(scatter1.get_values("wavedir"))

    print(scatter1.mean_xy())


if __name__ == "__main__":
    main()



from datetime import datetime
from typing import Sequence

from dateutil import rrule

from bluesmet.normet.dataset import Dataset

from ..combine import get_values_between as get_values


def get_values_between(
    lat_pos: float,
    lon_pos: float,
    start_date: datetime,
    end_date: datetime,
    requested_values: Sequence[str],
    cache_location = "./cache/zdepths_his/"
):
    """Return values for norkyst current subset"""

    return get_values(requested_values, lat_pos, lon_pos, "lon", "lat", start_date,end_date,rrule.DAILY,cache_location, __get_url)

def __get_url(date: datetime):
    # * https://thredds.met.no/thredds/fou-hi/fou-hi.html
    # ** https://thredds.met.no/thredds/fou-hi/norkyst800v2.html
    # *** https://thredds.met.no/thredds/catalog/fou-hi/norkyst800m-1h/catalog.html
    basename = "NorKyst-800m_ZDEPTHS_his.an"
    datlab = date.strftime("%Y%m%d")
    hour = "00"
    base_url = "https://thredds.met.no/thredds/dodsC/fou-hi/norkyst800m-1h"
    return f"{base_url}/{basename}.{datlab}{hour}.nc"


def get_metadata():
    """Get metadata"""
    start_date = datetime(2017, 2, 20)
    url = __get_url(start_date)
    ds = Dataset("./cache/zdepths_his/", url, start_date)
    metadata = ds.get_metadata()
    global_meta = {
        "fromDate": start_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
        "url": "https://thredds.met.no/thredds/catalog/fou-hi/norkyst800m-1h/catalog.html",
    }
    metadata["global"] = global_meta
    return metadata



import os
import shutil
from pathlib import Path
from datetime import datetime
from typing import Dict
import netCDF4 as nc

class Cache:

    def __init__(self, location):
        self.location = Path(location).absolute()

    def __store_subset(self, src,names, file: Path):
        pdir = Path(file).parent
        if not pdir.exists():
            os.makedirs(pdir)

        intermediate = file.with_suffix(".tmp")

        if file.exists():
            shutil.copy(file, intermediate)

        try:
            with nc.Dataset(intermediate, 'w', format='NETCDF4') as dst:
                # copy attributes
                for name in src.ncattrs():
                    dst.setncattr(name, src.getncattr(name))
                dimensions = {}
                for name in names:
                    variable = src.variables[name]
                    for dim in variable.dimensions:
                        dimensions[dim]=src.dimensions[dim]

                for name, dimension in dimensions.items():
                    dst.createDimension(name, len(dimension) if not dimension.isunlimited() else None)

                for name in names:
                    variable = src.variables[name]
                    x = dst.createVariable(name, variable.datatype, variable.dimensions)
                    x[:] = variable[:]

            shutil.move(intermediate, file)
        except Exception as e:
            os.remove(intermediate)
            raise e


    def get_root_variables(self,url: str, names, dimensions=None) -> Dict:
        """Get the time independent variables from the cache or from the source url"""
        if dimensions is None:
            dimensions = {}
        filename = self.location / "root.nc"
        variables = {}
        if not os.path.exists(filename):
            # Then we must transfer these values to the cache
            with nc.Dataset(url,"r") as f0:
                self.__store_subset(f0,names, filename)
                for name in names:
                    variable = f0.variables[name]
                    variables[name]=variable[:]
                    dimensions[name]=variable.dimensions
        else:
            with nc.Dataset(filename,"r") as f0:
                missing = [name for name in names if name not in f0.variables.keys()]
                if len(missing) == 0:
                    for name in names:
                        variable = f0.variables[name]
                        variables[name]=variable[:]
                        dimensions[name]=variable.dimensions
                else:
                    # There are missing variables, so we must transfer these values to the cache
                    with nc.Dataset(url,"r") as f0:
                        self.__store_subset(f0,names, filename)
                    # Now we can return the values
                    with nc.Dataset(filename,"r") as f0:
                        for name in names:
                            variable = f0.variables[name]
                            variables[name] = variable[:]
                            dimensions[name]=variable.dimensions
        return variables

    def to_cache_path(self, ddate: datetime,dim1: Dict, dim2: Dict) -> Path:
        """Convert a date and two dimensions to a path in the cache"""
        year = ddate.strftime("%Y")
        month = ddate.strftime("%m")
        day = ddate.strftime("%d")
        hour = ddate.strftime("%H")
        idx1 = next(iter(dim1.values()))
        idx2 = next(iter(dim2.values()))
        filename = f"{idx1},{idx2}.nc"
        return Path(f"{self.location}/{year}/{month}/{day}/{hour}/{filename}")



""" Python generator utilities """
import shutil
from pathlib import Path
from setuptools import setup

here = Path(__file__).parent.resolve()

# Remove build and dist folders
shutil.rmtree(Path("build"), ignore_errors=True)
shutil.rmtree(Path("dist"), ignore_errors=True)

setup()

shutil.rmtree(Path("build"), ignore_errors=True)


from datetime import datetime
from pathlib import Path
from typing import Callable, Sequence

import concurrent.futures
import pandas as pd
import xarray as xr
from dateutil import rrule

from .dataset import Dataset


def get_values_between(
    requested_values: Sequence[str],
    lat_pos: float,
    lon_pos: float,
    lon_name: str,
    lat_name: str,
    start_date: datetime,
    end_date: datetime,
    frequency: int,
    cache_location: str,
    url_converter: Callable[[datetime], str],
    max_concurrent_downloads = 1
):
    """Return values beween the given dates for the given coordinates"""

    coordinates = {
        "latitude": lat_pos,
        "longitude": lon_pos,
        "lon_name": lon_name,
        "lat_name": lat_name,
    }

    # The list of required values might be altered in lower layers if dependent values are required
    all_values = list(requested_values)

    dates = rrule.rrule(frequency, dtstart=start_date, until=end_date)
    paths = []

    def process_date(date):
        dimensions = {}
        path = __get_values_for_date(
            date, coordinates, all_values, dimensions, cache_location, url_converter
        )
        return path


    if max_concurrent_downloads > 1:
        # Use ThreadPoolExecutor for parallel processing
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_concurrent_downloads) as executor:
            # Submit tasks for each date and get the results as they complete
            future_to_date = {executor.submit(process_date, date): date for date in dates}
            # Wait for all to complete
            try:
                concurrent.futures.wait(future_to_date)
                for future in future_to_date:
                    path = future.result()
                    paths.append(path)
            # pylint: disable=broad-except
            except Exception as e:
                print(f"Error processing dates: {e}")
    else:
        # No concurrent processing
        for date in dates:
            path = process_date(date)
            paths.append(path)
        

    values = xr.Dataset()
    uninitialized = True

    # Open and combine the datasets
    for file_path in paths:
        ds = xr.open_dataset(file_path)
        if uninitialized:
            uninitialized = False
            values = ds
        else:
            values = xr.concat([values, ds], dim='time')
        ds.close()

    attributes = values.attrs
    attributes["latitude"] = coordinates["latitude_actual"]
    attributes["longitude"] = coordinates["longitude_actual"]

    # Now convert the time to datetime before we return the dataset
    values["time"] = pd.to_datetime(values["time"].values, unit='s')
    return values.sel(time=slice(start_date, end_date))


def __create_dataset(date: datetime, cache_location, url_converter):
    """Create dataset"""
    url = url_converter(date)
    ds = Dataset(cache_location, url, date)
    return ds


def __get_values_for_date(
    date: datetime,
    coordinates,
    requested_values,
    dimensions,
    cache_location,
    url_converter,
) -> Path:
    """Return values for nora3 wave subset"""
    ds = __create_dataset(date, cache_location, url_converter)
    return ds.get_values(coordinates, requested_values, dimensions)